@@ -42,8 +42,6 @@ static GeglRectangle get_bounding_box          (GeglOperation        *self);
 static GeglRectangle get_required_for_output   (GeglOperation        *operation,
                                                  const gchar         *input_pad,
                                                  const GeglRectangle *roi);
-static GeglRectangle  get_cached_region        (GeglOperation       *operation,
-                                                 const GeglRectangle *roi);
 
 
 static void
@@ -53,7 +51,6 @@ gegl_operation_source_class_init (GeglOperationSourceClass * klass)
 
   operation_class->process = gegl_operation_source_process;
   operation_class->attach  = attach;
-  operation_class->get_cached_region = get_cached_region;
 
   operation_class->get_bounding_box  = get_bounding_box;
   operation_class->get_required_for_output = get_required_for_output;
@@ -207,10 +204,3 @@ get_required_for_output (GeglOperation        *operation,
 {
   return *roi;
 }
-
-static GeglRectangle
-get_cached_region (GeglOperation       *operation,
-                   const GeglRectangle *roi)
-{
-  return *roi;
-}
